package ir.classifiers;

import java.io.*;
import java.util.*;
import ir.vsr.*;
import ir.utilities.*;

/**
 * Gives learning curves with K-fold cross validation for a classifier.
 *
 * @author       Sugato Basu
 */

public class CVLearningCurve
{
  /** Stores the possible class labels */
  public static final String[] CLASSES = {"bio","chem","phys"};
  /** Points at which the learning curve is plotted */
  public static final double[] POINTS = {0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,1};
  /** Stores all the examples for each class */
  protected Vector [] totalExamples;
  /** foldBins[i][j] stores the examples for class i in fold j. This stores the training-test splits for all the folds */
  protected Vector [][] foldBins;
  /** The classifier for which K-fold CV learning curve has to be generated */
  protected Classifier classifier;
  /** Number of classes in the data */
  protected int numClasses;
  /** Number of folds of cross validation to run */
  protected int numFolds;
  /** Flag for debug display */
  protected boolean debug=false;
  /** Total Training time */
  public double trainTime;
  /** Total Testing time */
  public double testTime;
  
  /** Stores a static InvertedIndex for classifiers that need it, 
   * for efficient memory management */
  private static InvertedIndex invertedIndex = null;
  
  /** Creates a CVLearning curve object
   *
   * @param nfolds Number of folds of CV to perform
   * @param c Classifier on which to perform K-fold CV
   * @param nclasses Number of classes (categories) in data
   * @param dirName Directory where all data files are present
   * @param d Flag to set elaborate display option
   */
  CVLearningCurve(int nfolds, Classifier c, int nclasses, String dirName, boolean d) {
    numFolds = nfolds;
    classifier = c;
    numClasses = nclasses;
    totalExamples = new Vector[numClasses];
    foldBins = new Vector[numClasses][numFolds];
    System.out.println("\nReading in documents from directory: " + dirName);
    setTotalExamples(dirName);
    debug = d;
    trainTime = testTime = 0;
  }
  
  /** Return classifier */
  public Classifier getClassifier() {
    return classifier;
  }
  
  /** Set the classifier */
  public void setClassifier(Classifier c) {
    classifier = c;
  }
  
  /** Return all the examples */
  public Vector [] getTotalExamples() {
    return totalExamples;
  }
  
  /** Set all the examples */
  public void setTotalExamples (Vector [] data) {
    totalExamples = data;
  }

  /** Return the fold Bins */
  public Vector [][] getFoldBins() {
    return foldBins;
  }
  
  /** Set the fold Bins */
  public void setFoldBins (Vector [][] bins) {
    foldBins = bins;
  }
  
  /** Sets the totalExamples by reading in file from directory dirName */
  public void setTotalExamples(String dirName) {
    DocumentIterator docIter = new DocumentIterator(new File(dirName), DocumentIterator.TYPE_HTML, false);
    while(docIter.hasMoreDocuments()) { //read in all documents
      FileDocument doc = docIter.nextDocument();
      int category = findClassID(doc.file.getName()); // find category of document
      Example example = new Example (doc.hashMapVector(), category, doc.file.getName(), doc);
      if(totalExamples[category] == null)
	totalExamples[category] = new Vector();
      totalExamples[category].add(example);
    }
  }
  
  /**
   * Generate a vector of predictions ready for processing, by performing a
   * cross-validation on the supplied dataset.
   *
   * @return The vector of predictions to be processed for plot generation
   *         Each component in the prediction vector is a vector of K
   *         [training set size, accuracy] values, corresponding to the K accuracy
   *         measurements at the point "size" on the learning curve
   */
  public Vector getCVPredictions() {
    Vector predictions = new Vector();
    
    // randomly mix the training examples in each category
    randomizeOrder();
    // create foldBins from totalExamples -- effectively creates the
    // training-test splits for each fold
    binExamples();
    
    for (int i = 0; i < POINTS.length; i++) {
      double percent = POINTS[i];
      System.out.println("\nPercentage: " + percent);
      Vector resultsPerPoint = new Vector();
      for (int fold = 0; fold < numFolds; fold++) {
	System.out.println("\nCalculating results for fold: " + fold);
	// Creates training data for this fold, from the first
	// percent % data in each of the training folds
	Vector train = getTrainCV(fold, percent);
	// Creates testing data for this fold
	Vector test = getTestCV(fold);
	// Get testing results for this fold and percent setting
	Vector foldPred = getTestPrediction(train, test);
	if (debug) {
	  System.out.println("Training on:\n" + train);
	  System.out.println("Testing on:\n" + test);
	}
	// add [training set size,accuracy] result for each fold
	resultsPerPoint.add(foldPred);
      }
      // add the vector of K [training set size, accuracy] values for one point on training curve
      predictions.add(resultsPerPoint);
    }
    System.out.println("\nFinished calculating all prediction vectors...");
    return predictions;
  }
  
  /**
   * Generate a prediction vector by performing an evaluation on the test set after 
   * training on the given training set.
   *
   * @param train The training dataset vector
   * @param test The testing dataset vector
   * @return The vector of prediction on test data, each prediction vector containing
   *         K [training set size, accuracy] vectors for each fold
   */
  public Vector getTestPrediction(Vector train, Vector test) {
    int correct = 0;
    Vector prediction = new Vector();
    
    long startTime = System.currentTimeMillis();
    // train the classifier on train data
    if(classifier.usesInvertedIndex() == true) { // for classifiers that use Inverted Indices
      if(invertedIndex == null) {
	invertedIndex = new InvertedIndex(train, DocumentIterator.TYPE_HTML,false,false);
      }
      else {
	// use existing inverted index for efficient memory management
	invertedIndex.docRefs.clear();
	invertedIndex.tokenHash.clear();
	invertedIndex.indexDocuments(train);
      }
      classifier.setInvertedIndex(invertedIndex);
    }
    classifier.train(train);
    double timeTaken = System.currentTimeMillis() - startTime;
    trainTime += timeTaken;
    
    startTime = System.currentTimeMillis();
    for (int i = 0; i < test.size(); i++) {
      Example example = (Example) test.get(i);
      // classify the test example
      if(classifier.test(example))
	correct++;
    }
    timeTaken = System.currentTimeMillis() - startTime;
    testTime += timeTaken;
    prediction.add(new Integer (train.size()));
    prediction.add(new Double (1.0*correct/test.size()));
    
    return prediction;
  }

    /** Set the fold Bins from the total Examples -- this effectively
     *  stores the training-test split
     */
    public void binExamples() {
      for (int classNum=0; classNum<numClasses; classNum++) {
	for (int j=0; j<totalExamples[classNum].size(); j++) {
	  int foldNum = 0;
	  if (numFolds > 1)
	    foldNum = j % numFolds;
	  if(foldBins[classNum][foldNum] == null)
	    foldBins[classNum][foldNum] = new Vector();
	  foldBins[classNum][foldNum].add(totalExamples[classNum].get(j));
	}
      }
    }

    /**
     * Creates the training set for one fold of a cross-validation 
     * on the dataset.
     *
     * @param foldnum The fold for which training set is to be constructed
     * @param percent Percentage of examples to use for training in this fold
     * @return The training data
     */
    public Vector getTrainCV(int foldnum, double percent) {
	Vector train = new Vector();
	
	for (int i=0; i<numClasses; i++) {
	    for (int j=0; j<numFolds; j++) {
		if(j!=foldnum) {
		    int max = (int) (percent*foldBins[i][j].size());
		    for(int k=0; k<max; k++) {
			train.add(foldBins[i][j].get(k));
		    }
		}
	    }
	}
	return train;
    }

    /**
     * Creates the testing set for one fold of a cross-validation 
     * on the dataset.
     *
     * @param foldnum The fold which is to be used as testing data
     * @return The test data
     */
    public Vector getTestCV(int foldnum) {
	Vector test = new Vector();
	
	for (int i=0; i<numClasses; i++)
	    test.addAll(foldBins[i][foldnum]);

	return test;
    }

    /** Finds the class ID from the name of the document */
    public static int findClassID (String name) {
	int ret = -1;
        if (name.indexOf(CLASSES[0]) != -1)
            ret = 0;
	if (name.indexOf(CLASSES[1]) != -1)
            ret = 1;
        else if (name.indexOf(CLASSES[2]) != -1)
            ret = 2;
        return ret;
    }

    /**
     * Shuffles the examples in totalExamples so that they are ordered randomly.
     */
    private final void randomizeOrder() {
	
	Random random = new Random();
	for (int i=0; i<numClasses; i++) {
	    int maxSize = totalExamples[i].size();
	    for (int j=maxSize-1; j>0; j--) {
		int next = random.nextInt(maxSize);
		Example temp = (Example) totalExamples[i].get(j);
		totalExamples[i].set(j, totalExamples[i].get(next));
		totalExamples[i].set(next, temp);
	    }
	}
    }

    /** Write out the final learning curve data.
     *  One line for each value: [training set size, accuracy]
     *  This is the format needed for GNUPLOT.
     *
     *  @param Vector of results from which GNUPLOT data is generated
     *  @name  Name of classifier
     */
    void writeCurve(Vector result, String name) throws IOException{
	PrintWriter out = new PrintWriter(new FileWriter(name + ".data"));
	for(int i=0; i<result.size(); i++) {
	    double accuracy = 0;
	    Vector entry = (Vector) result.get(i);
	    int size = ((Integer) ((Vector) entry.get(0)).get(0)).intValue();
	    for (int j=0; j<entry.size(); j++) {
		accuracy += ((Double) ((Vector) entry.get(j)).get(1)).doubleValue();
	    }
	    // find average accuracy across the K folds
	    accuracy /= entry.size();
	    out.println(size + "\t" + accuracy);
	}
	out.close();
    }

    /** Write out an appropriate input file for GNUPLOT for the final 
     *  learning curve  to the output file with a ".gplot" extension.
     *  See GNUPLOT documentation. 
     *
     *  @param Vector of results from which GNUPLOT data is generated
     *  @name  Name of classifier
     */
    void makeGnuplotFile(Vector result, String name) throws IOException{
	writeCurve(result, name);
	File graphFile = new File(name + ".gplot");
	PrintWriter out = new PrintWriter(new FileWriter(graphFile));
	out.print("set xlabel \"Size of training set\"\nset ylabel \"Accuracy\"\n\nset terminal postscript color\nset size 0.75,0.75\n\nset data style linespoints\n\nplot \'" + name + ".data\' title \"" + name + ": 10-fold CV Learning Curve\"");
	out.close();
    }
}
