package ir.eval;

import java.io.*;
import java.util.*;
import java.lang.*;
import ir.utilities.*;
import ir.vsr.*;

/**
 * A specialization of Experiment for evaluating relevance feedback.
 * This requires, first determining the top ranked retrievals, then simulating
 * relevance feedback using known relevant docs from the test data. However,
 * performance must then be evaluated on only documents for which explicit
 * relevance information was NOT provided.
 *
 * @author Ray Mooney
 */

public class ExperimentRelFeedback extends Experiment {

    /** Whether stemming is used in the experiment. */
    boolean stem;

    /** The number of top ranked documents to simulate relevance feedback for. */
    int numFeedback;

    /** Whether to actually use feedback, or just run as a control to test normal
     * perfomance without the numFeedback top ranked documents in the test set */
    boolean feedback;

    /** The number of top ranked examples not to use for testing (performance evaluation).
     * This is usually the same as numFeedback but may be greater if one wants to test
     * the effect of different amounts of feedback on the same test corpus.
     * Note that (feedback = false and numFeedBack = numSkip = n) is the same as 
     * (feedback = true, numFeedback = 0, and numSkip = n) */
    int numSkip;

    /** Create an experiment for relevance feedback with the given parameters */    
    public ExperimentRelFeedback(File corpusDir, File queryFile, File outFile, short docType, 
		      boolean stem, int numFeedback, boolean feedback, int numSkip) throws IOException {
	super(corpusDir, queryFile, outFile, docType, stem);
	this.stem = stem;
	this.numFeedback = numFeedback;
	this.feedback = feedback;
	this.numSkip = numSkip;
    }

    /** Specialized version for processing a query and simulating reelvance feedback */
    boolean processQuery(BufferedReader in) throws IOException {
	String query = in.readLine();   // get the query
	if (query == null) return false;  // return false if end of file
	System.out.println("\nQuery " + (rpResults.size() + 1) + ": " + query);
	// Process the query and get the ranked retrievals
	HashMapVector queryVector = (new TextStringDocument(query,stem)).hashMapVector();
	Retrieval[] retrievals = index.retrieve(queryVector);
	System.out.println("Returned " + retrievals.length + " documents.");
	// Read the known relevant docs from query file and parse them
	// into an ArrayList of String's of relevant file names.
	String line = in.readLine();
	ArrayList correctRetrievals = MoreString.segment(line,' ');
	System.out.println(correctRetrievals.size() + " truly relevant documents.");

	// Create a relevance feedback object for the given query
	Feedback fdback = new Feedback(queryVector, retrievals, index);

	// For the top numFeedback examples in the retrieved documents,
	for(int i = 0; i < Math.min(numFeedback,retrievals.length); i++) {
	    // If the document is in the set of correct retrievals for this query
	    if (correctRetrievals.contains(retrievals[i].docRef.file.getName()))
		// Simulate it as positive relevance feedback
		fdback.addGood(retrievals[i].docRef);
	    else
		// Simulate it as negative relevance feedback
		fdback.addBad(retrievals[i].docRef);
	}
	System.out.println("Feedback:\nPositive docs: " + fdback.goodDocRefs + 
			   "\nNegative docs: " + fdback.badDocRefs);

	// Determine the List of documents NOT to include in the test set
	List skipDocs = new ArrayList(numSkip);
	// Include in this set the first numSkip docs in the list of
	// initial retrievals.
	for(int i = 0; i < Math.min(numSkip,retrievals.length); i++) {
	    skipDocs.add(retrievals[i].docRef);
	}
	if (feedback) {
	    // Determine a new query using the simulated feedback 
	    System.out.println("Executing New Expanded and Reweighted Query: ");
	    queryVector = fdback.newQuery();
	    // Get the new set of retrievals
	    retrievals = index.retrieve(queryVector);
	    // Update the list of retrievals stored in the feedback
	    fdback.retrievals = retrievals;
	}
	
	// Remove the documents to be skipped from the set of correctRetrievals 
	// and new retrievals,
	removeSkipDocs(correctRetrievals, skipDocs);
	retrievals = removeSkipDocs(retrievals, skipDocs);

	// Generate Recall/Precision points and save in rpResults
	rpResults.add(evalRetrievals(retrievals, correctRetrievals));
	// Read the blank line delimiter between queries in the query file
	line = in.readLine();
	if (! (line == null || line.trim().equals(""))) {
	    System.out.println("\nCould not find blank line after query, bad queryFile format");
	    System.exit(1);
	}
	return true;
    }

    /** Remove the skipDocRefs from the correctRetrievals */
    void removeSkipDocs(List correctRetrievals, List skipDocRefs) {
	// Construct a list of the file names of all the skipDocRefs
	ArrayList skipFileNames = new ArrayList();
	for(int i = 0; i < skipDocRefs.size(); i++) {
	    skipFileNames.add(((DocumentReference)skipDocRefs.get(i)).file.getName());
	}
	// Remove all these from the list of file names in correctRetrievals
	correctRetrievals.removeAll(skipFileNames);
    }

    /** Remove the skipDocRefs from the current retrievals 
     @return the new set of retrievals */
    Retrieval[] removeSkipDocs(Retrieval[] retrievals, List skipDocRefs) {
	// Create an array for the reduced result
	Retrieval[] result = new Retrieval[retrievals.length - skipDocRefs.size()];
	// Count is the number of retrievals currently retained and added to result
	int count = 0;
	// Check wether to retain each retreival
	for(int i = 0; i < retrievals.length; i++) {
	    DocumentReference docRef = retrievals[i].docRef;
	    // If the doc is not one to be skipped
	    if (!skipDocRefs.contains(docRef)) {
		if (count >= result.length) {
		    System.out.println("\nCould not find all skipDocs in retrievals");
		    System.exit(1);
		}
		// Add it to be the next entry in the result
		result[count] = retrievals[i];
		count++;
	    }
	}
	// Should just fill the result with all non-skipped documents
	if (count != result.length) {
	    System.out.println("\nFound too many skipDocs in retrievals");
	    System.exit(1);
	}
	return result;
    }

    /** Evaluate retrieval preformance on a given query test corpus and
     * generate a recall/precision graph.
     * Command format: "Experiment [OPTION]* [DIR] [QUERIES] [OUTFILE] [FEEDBACKNUM]"  where:
     * DIR is the name of the directory whose files should be indexed.
     * QUERIES is a file of queries paired with relevant docs (see queryFile).
     * OUTFILE is the name of the file to put the output. The plot
     * data for the recall precision curve is stored in this file and a
     * gnuplot file for the graph is the same name with a ".gplot" extension.
     * FEEDBACKNUM is the number of top retrievals to simulate feedback for.
     * OPTIONs can be
     * "-html" to specify HTML files whose HTML tags should be removed, and
     * "-stem" to specify tokens should be stemmed with Porter stemmer.
     * "-control" to not actually do the feedback but evaluate performance
     *     for no feedback on the same test docs (eliminating the feeback examples)
     * "-skip N" where N is an int, to skip the top N ranked docs when
     *      evaluating final performance.  By default, the number skipped is just
     *      FEEDBACKNUM, but N can be greater than FEEDBACKNUM if want to
     *      evaluate the performance of different levels of feedback on the 
     *      same test docs, using the maximum FEEDBACKNUM as the N for -skip (numSkip)
     *      for different runs with different FEEDBACKNUMs.
     *      Note that "-skip N" with  FEEDBACKNUM = 0 is the same as
     *      "-control" with FEEDBACKNUM = N
     */
    public static void main(String[] args) throws IOException {
	// Parse the arguments into a directory name and optional flag
	String corpusDir = args[args.length - 4];
	String queryFile = args[args.length - 3];
	String outFile = args[args.length - 2];
	int numFeedback = Integer.parseInt(args[args.length - 1]);
	boolean feedback = true;
	short docType = DocumentIterator.TYPE_TEXT;
	boolean stem = false;
	int numSkip = numFeedback;
	for(int i = 0; i < args.length - 4; i++) {	
	   String flag = args[i];
	   if (flag.equals("-html"))
		// Create HTMLFileDocuments to filter HTML tags
		docType = DocumentIterator.TYPE_HTML;
	   else if (flag.equals("-stem"))
	       // Stem tokens with Porter stemmer
	       stem = true;
	   else if (flag.equals("-control"))
	       // Stem tokens with Porter stemmer
	       feedback = false;
	   else if (flag.equals("-skip"))
	       numSkip = numFeedback;
	   else if (args[i - 1].equals("-skip")) {
	       numSkip = Integer.parseInt(flag);
	       if (numSkip < numFeedback) {
		   System.out.println("\nMust skip more than numFeedback: " + flag);
		   System.exit(1);
	       }
	   }
	   else {
	       System.out.println("\nUnknown flag: " + flag);
	       System.exit(1);
	   }
	}
	if (!feedback && numSkip != numFeedback) {
	    System.out.println("\n-control with -skip, not sensible");
	    System.exit(1);
	}
	ExperimentRelFeedback exper = new ExperimentRelFeedback(new File(corpusDir), new File(queryFile), 
								new File(outFile), docType, stem, 
								numFeedback, feedback, numSkip);
        exper.makeRpCurve();
    }

}
