package ir.vsr;

import java.io.*;
import java.util.*;
import ir.utilities.*;

/**
 * An inverted index for vector-space information retrieval. Contains
 * methods for creating an inverted index from a set of documents
 * and retrieving ranked matches to queries using standard TF/IDF
 * weighting and cosine similarity, but also accounts for "proximity"
 * of the terms in a document.
 *
 * Documents are ranked by the ratio of normal vector-similarity
 * and a distance measure that captures how far apart the query terms
 * appear in the document.  The distance (proximity) measure is based
 * on the average across all pairs of terms in the query, of the average
 * distance in the document between the one word in the pair and the closest
 * occurrence of the other word in the pair.  Pairs appearing in the 
 * wrong order relative to the query get an additional distance penalty.
 *
 * @author Ray Mooney
 */

public class InvertedPosIndex extends InvertedIndex
{
    /** The maximum measurable distance that can separate two query words. */
    public static final double MAX_DISTANCE = 1000;
    /** The multiplicative penalty factor for distance that is incurred when
	query terms are in the opposite order in the document */
    public static final double WRONG_ORDER_PENALTY_FACTOR = 2;

    public InvertedPosIndex(File dirFile, short docType, boolean stem) {
	super(dirFile, docType, stem, false);
    }

    /** Index the documents in dirFile. */
    protected void indexDocuments() {
	if (!tokenHash.isEmpty() || !docRefs.isEmpty()) {
	    // Currently can only index one set of documents when an index is created
	    System.out.println("\nCannot indexDocuments more than once in the same InvertedIndex");
	    System.exit(1);
	}	    
	// Get an iterator for the documents
	DocumentIterator docIter = new DocumentIterator(dirFile, docType, stem);
	System.out.println("Indexing documents in " + dirFile);
	// Loop, processing each of the documents
	while (docIter.hasMoreDocuments()) {
	    FileDocument doc = docIter.nextDocument();
	    // Create a document positional vector for this document
	    HashMapPosVector vector = doc.hashMapPosVector();
	    // Create a reference to this document
	    DocumentReference docRef = new DocumentReference(doc);
	    // Add this document to the list of documents indexed
	    docRefs.add(docRef);
	    // Iterate through each of the tokens in the document
	    Iterator mapEntries = vector.iterator();
	    while (mapEntries.hasNext()) {
		Map.Entry entry = (Map.Entry)mapEntries.next();
		// An entry in the HashMap maps a token to an ArrayList of token positions 
		String token = (String)entry.getKey();
		// Get the position list for the token 
		ArrayList positions = (ArrayList)entry.getValue();
		// Add an occurence of this token to the inverted index pointing to this document
		indexToken(token, positions, docRef);
	    }
	}
	// Now that all documents have been processed, we can calculate the IDF weights for
	// all tokens and the resulting lengths of all weighted document vectors.
	computeIDFandDocumentLengths();
	System.out.println("Indexed " +  docRefs.size() + " documents with " + size() + " unique terms.");
    }

    /** Add a token occurrence to the index.
     * @param token The token to index.
     * @param positions An ArrayList of Integer positions of the token in the doc
     * @param docRef A reference to the Document it occurs in.
     */
    protected void indexToken(String token, ArrayList positions, DocumentReference docRef) {
	// Find this token in the index
	TokenInfo tokenInfo = (TokenInfo)tokenHash.get(token);
	if (tokenInfo == null) {
	    // If this is a new token, create info for it to put in the hashtable
	    tokenInfo = new TokenInfo();
	    tokenHash.put(token, tokenInfo);
	}
	// Add a new occurrence for this token to its info
	tokenInfo.occList.add(new TokenPosOccurrence(docRef, positions));
    }

    /** TokenOccurence in an InvertedPosIndex should be a TokenPosOccurrence, so print
     * the positional information for this occurrence. */
    protected void printExtraTokenOccurrenceInfo(TokenOccurrence occ) {
	TokenPosOccurrence posOcc = (TokenPosOccurrence)occ;
	System.out.print("; Pos: [");
	for(int i = 0; i < posOcc.positions.length; i++) {	
	    System.out.print(posOcc.positions[i]);
	    if (i < (posOcc.positions.length - 1))
		System.out.print(",");
	}
	System.out.println("]");
    }

    /** Perform ranked retrieval on this input query Document. */
    public Retrieval[] retrieve(Document doc) {
	return retrieve(doc.positionOrderedTokenVector());
    }

    /** Perform ranked retrieval on this input query Document vector. */
    public Retrieval[] retrieve(TokenPositionInfo[] vector) {
	// Create a hashtable to store the retrieved documents.  Keys
	// are docRefs and values are DoubleValues which indicate the
	// partial score accumulated for this document so far.
	// As each token in the query is processed, each document
	// it indexes is added to this hashtable and its retrieval
	// score (similarity to the query) is appropriately updated.
	HashMap retrievalHash = new HashMap();
	// Initialize a variable to store the length of the query vector
	double queryLength = 0.0;
	// Iterate through each token in the query input Document in the
	// order in which they first occur
	for(int i = 0; i < vector.length; i++) {	
	    // Get the token and the count for each token in the query
	    String token = vector[i].token;
	    int count = vector[i].positions.size();
	    //	    System.out.println("Word " + i + ": " + token + " Count: " + count);
	    // Determine the score added to the similarity of each document
	    // indexed under this token and update the length of the
	    // query vector with the square of the weight for this token.
	    queryLength = queryLength + incorporateToken(token, count, retrievalHash);
	}
	// Finalize the length of the query vector by taking the square-root of the
	// final sum of squares of its token wieghts.
	queryLength = Math.sqrt(queryLength);
	// Make an array to store the final ranked Retrievals.
	Retrieval[] retrievals = new Retrieval[retrievalHash.size()];
	// Iterate through each of the retrieved documents stored in
	// the final retrievalHash.
	Iterator rmapEntries = retrievalHash.entrySet().iterator();
	int retrievalCount = 0;
	while (rmapEntries.hasNext()) {
	    // Get the DocumentReference and RetrievalPosInfo for each retrieved document
	    Map.Entry entry = (Map.Entry)rmapEntries.next();
	    DocumentReference docRef = (DocumentReference)entry.getKey();
	    RetrievalPosInfo info = (RetrievalPosInfo)entry.getValue();
	    double score = info.vectorScore;
	    // Normalize score for the lengths of the two document vectors
	    score = score / (queryLength * docRef.length);  
	    // Finalize the final average for the proximity score
	    finalizeProximityScore(info, vector.length);
	    // The final score for a document is the vector-space
	    // score divided by the proximity score scaled by the
	    // the MAX_DISTANCE
	    if (info.proximityScore != 0)
		score = score / (info.proximityScore / MAX_DISTANCE);
	    // Add a Retrieval for this document to the result array
	    retrievals[retrievalCount++] = new Retrieval(docRef, score);
	}
	// Sort the retrievals to produce a final ranked list using the
	// Comparator for retrievals that produces a best to worst ordering.
	Arrays.sort(retrievals);
	return retrievals;
    }

    /** Retrieve the documents indexed by this token in the inverted index,
     * add it to the retrievalHash if needed, and update its running scores.
     * @param token The token in the query to incorporate.
     * @param count The count of this token in the query.
     * @param retrievalHash The hashtable of retrieved DocumentReferences and related RetrievalPosInfo
     * @return The square of the weight of this token in the query vector for use
     * in calculating the length of the query vector.
     */
    public double incorporateToken(String token, int count, HashMap retrievalHash) {
	TokenInfo tokenInfo = (TokenInfo)tokenHash.get(token);
	// If token is not in the index, it adds nothing and its squared weight is 0
	if (tokenInfo == null) return 0.0;
	// The weight of a token in the query is is IDF factor times the number
	// of times it occurs in the query.
	double weight = tokenInfo.idf * count;
	// For each document occurrence indexed for this token...
	for(int i = 0; i < tokenInfo.occList.size(); i++) {
	    // Get the ith document occurrence indexed under this token
	    TokenPosOccurrence occ = (TokenPosOccurrence)tokenInfo.occList.get(i);
	    //	    System.out.println("Found in " + occ.docRef.file + " " + occ.count + " times.");
	    // Get the current  RetrievalPosInfo for this document in the retrievalHash.
	    RetrievalPosInfo info = (RetrievalPosInfo)retrievalHash.get(occ.docRef);
	    if (info == null) {
		// If this is a new retrieved document, create an initial RetrievalPosInfo
		// for it and store in the retrievalHash
		info = new RetrievalPosInfo();
		retrievalHash.put(occ.docRef, info);
	    }
	    // Update the vector space score for this document by adding the product
	    // of the weight of this token in the query and its weight
	    // in the retrieved document (IDF * occurrence count)
	    info.vectorScore = info.vectorScore + weight * tokenInfo.idf * occ.count;
	    // Update the proximity score for the RetrievalPosInfo for this retrieved document
	    updateProximityScore(info, occ, token);
	    // Add the current query token to the list of found tokens and its position vector in the RetrievalPosInfo
	    info.foundTokens.add(token);
	    info.foundPositions.add(occ.positions);
	}
	// Return the square of the weight of this token in the query
	return weight*weight;
    }

    /** Update the proximity score of the RetrievalPosInfo of this retrieved document based on 
     * how close the current query token appears to each of the previous query tokens found in this
     * document.
     */
    protected static void updateProximityScore(RetrievalPosInfo info, TokenPosOccurrence occ, String token) {
	// For each previously found query token, 
	for(int i = 0; i < info.foundPositions.size(); i++) {
	    // add to the proximity score the average closest distance of the current query token to this
	    // previously found query token for this document
	    double dist = averageClosestDistance((int[])info.foundPositions.get(i), occ.positions);
	    //	    System.out.println("Avg Dist between " + token + " and " + info.foundTokens.get(i) + " = " + dist);
	    info.proximityScore = info.proximityScore + dist;
	}
    }

    /** Finalize the proximity score for a RetrievalPosInfo by averaging over all possible pairs of
     * query tokens, the average closest distance score for each pair 
     */
    protected static void finalizeProximityScore(RetrievalPosInfo info, int queryLength) {
	// No proximity score for single word query
	if (queryLength == 1 ) return;
	// The total number of pairs of query terms
	int numPairs = queryLength * (queryLength - 1) / 2;
	// The number of pairs of query terms that were actually found in the document and therefore for whom
	// the proximity score was updated with their average closest distance in updateProximityScore
	int countedPairs = info.foundTokens.size() * (info.foundTokens.size() - 1) / 2;
	// The number of uncounted pairs due to unfound query terms (by elimination)
	int uncountedPairs = numPairs - countedPairs;
	// Assume the score for each uncounted pair is MAX_DISTANCE and compute an average over
	// all pairs of query terms.
	info.proximityScore = (info.proximityScore + uncountedPairs * MAX_DISTANCE) / numPairs;
    }

    /** Returns the average closest positional distance between an occurrence of the current query token and 
     * an occurrence of a specified previous query token.  If the tokens appear in the wrong order in the text
     * (i.e. the previous token appears after the current token), there is a penalty factor of 
     * WRONG_ORDER_PENALTY_FACTOR multiplied by the distance.  The "closest" occurrence taking 
     * into account this penalty is what is counted in computing the average closest distance.
     * @param previousPositions The locations of the previous query token in the text.
     * @param positions The locations of the current query token in the text.
     */
    protected static double averageClosestDistance(int[] previousPositions, int[] positions) {
	// A running total of the summed closest distances in order to calculate the average at the end
	double sumDistance = 0;
	// For each position of the current query token, find the closest occurrence (both preceding and
	// following) of the previous query token.
	for(int i = 0; i < positions.length; i++) {
	    // Use binary search to find the location where the current position should be
	    // "inserted" in the ordered list of previous token positions in order to efficiently locate
	    // the closest preceding and following occurences of the previous query token relative to the current
	    // query token.
	    int index = Arrays.binarySearch(previousPositions, positions[i]);
	    if (index >= 0) {
		// It should never actually find the current token position in the list of previous
		// token positions since they are always distinct tokens in the query vector
		System.out.println("Found token in same position as previous token, should not happen");
		System.exit(1);
	    }
	    // The insertion point is calculated based on the value returned by Arrays.binarySearch
	    // (see the javadoc for this method).  Works since previousPositions vector is sorted in 
	    // position order. The position of the closest previous occurrence of the previous token should
	    // then be in previousPositions[insertionPoint - 1] and the position of the closest following 
	    // occurrence  of the previous token should be previousPositions[insertionPoint].
	    int insertionPoint = 0 - index - 1;

	    // Closest distanceBefore and distanceAfter are initialized to -1 as an indication of "unset"
	    int distanceBefore = -1, distanceAfter = -1;

	    // Calculate the distance to the closest prior appearance of the previous token, if it exists
	    if (insertionPoint != 0) {
		distanceBefore = positions[i] - previousPositions[insertionPoint - 1];
		if (distanceBefore <= 0) {
		    System.out.println("Negative token distance, should not happen");
		    System.exit(1);
		}
	    }
	    // Calculate the distance to the closest following appearance of the previous token, if it exists.
	    if (insertionPoint < previousPositions.length) {
		distanceAfter = previousPositions[insertionPoint] - positions[i];		
		if (distanceAfter <= 0) {
		    System.out.println("Negative token distance, should not happen");
		    System.exit(1);
		}
	    }
	    double distance; 
	    if (distanceBefore == -1)
		// There is no correctly ordered occurrence of the two tokens, just an incorrect one
		distance = distanceAfter * WRONG_ORDER_PENALTY_FACTOR;
	    else if (distanceAfter == -1)
		// There is no incorrectly ordered occurrence of the two tokens, just a correct one
		distance = distanceBefore;
	    else
		// Return the minimum of the distance score for the closest correctly ordered
		// and incorrectly ordered occurrences
		distance = Math.min(distanceBefore, distanceAfter * WRONG_ORDER_PENALTY_FACTOR);
	    // Update the total sum of closest distances in order to keep track of the average
	    sumDistance = sumDistance + distance;
	}
	// Return the minimum of the average closest distance and the MAX_DISTANCE
	return Math.min(sumDistance/positions.length, MAX_DISTANCE);
    }

    /** Enter an interactive user-query loop, accepting queries and showing the retrieved 
     * documents in ranked order. 
     */
    public void processQueries() {

	System.out.println("Now able to process queries. When done, enter an empty query to exit.");
	// Loop indefinitely answering queries
	do {
	    // Get a query from the console
	    String query = UserInput.prompt("\nEnter query:  ");
	    // If query is empty then exit the interactive loop
	    if (query.equals(""))
		break;
	    // Get the ranked retrievals for this query string and present them.
	    // First get an ordered array of query tokens for retrieve.
	    TokenPositionInfo[]	queryTokens = (new TextStringDocument(query,stem)).positionOrderedTokenVector();
	    Retrieval[] retrievals = retrieve(queryTokens);
	    presentRetrievals(null, retrievals);
	}
	while(true);
    }

    /** Index a directory of files and then interactively accept retrieval queries.
     * Command format: "InvertedIndex [OPTION] [DIR]" where DIR is the name of
     * the directory whose files should be indexed, and the optional OPTION
     * can be "-html" to specify HTML files whose HTML tags should be removed.
     * and "-stem" to specify Porter stemming of tokens.
     */
    public static void main(String[] args) {
	// Parse the arguments into a directory name and optional flag
	String dirName = args[args.length - 1];
	short docType = DocumentIterator.TYPE_TEXT;
	boolean stem = false;
	for(int i = 0; i < args.length - 1; i++) {	
	   String flag = args[i];
	   if (flag.equals("-html"))
		// Create HTMLFileDocuments to filter HTML tags
		docType = DocumentIterator.TYPE_HTML;
	   else if (flag.equals("-stem"))
	       // Stem tokens with Porter stemmer
	       stem = true;
	   else {
	       System.out.println("\nUnknown flag: " + flag);
	       System.exit(1);
	   }
	}
	// Create an inverted index for the files in the given directory.
	InvertedPosIndex index = new InvertedPosIndex(new File(dirName), docType, stem);
	// index.print();
	// Interactively process queries to this index.
	index.processQueries();
    }


}

