package ir.webutils;

import java.util.*;
import java.net.*;
import java.io.*;
import ir.utilities.*;

public class AnchoredSpider extends Spider {

    /** Map from URLs to file names where they are stored.
     * This allows attaching anchor text of links to this
     * page to the bottom of the page. */
    protected HashMap urlMap;
    
    /**
     * Called when "-u" is passed in on the command line.  <p> This
     * implementation adds <code>value</code> to the list of links to
     * visit.  This version creates an initial <b>anchored</b> link.
     *
     * @param value The value associated with the "-u" option. */
    protected void handleUCommandLineOption(String value) {
	linksToVisit.add(new AnchoredLink(value));
    }

    /** This crawl differs from the default by extracting
     * {@link AnchoredLink}'s and attaching all anchor text
     * to the bottom of indexed pages to which they point.
     */
    public void doCrawl() {
	urlMap = new HashMap();
	while (linksToVisit.size() > 0 && count < maxCount) {
	    // Pause if in slow mode
	    if (slow) {
		synchronized (this) {
		    try {
			wait(1000);
		    }
		    catch (InterruptedException e) {
		    }
		}
	    }
	    // Take the top link off the queue
	    AnchoredLink link = (AnchoredLink) linksToVisit.remove(0);
	    System.out.println("Trying: " + link);
	    String existingFile = (String)urlMap.get(link);
	    if (existingFile == null) 
		// If new URL, store it in the urlMap, initially with
		// an empty file name for the value. Later replace value 
		// with file name if it is indexed.
		urlMap.put(link, "");
	    else {
		System.out.println("Already visited");
		// If URL was stored in a file, add the new anchor text
		if (!existingFile.equals("")) 
		    addAnchorText(existingFile, link, false);
		continue;
	    }
	    if (!linkToHTMLPage(link)) {
		System.out.println("Not HTML Page");
		continue;
	    }
	    HTMLPage currentPage = null;
	    // Use the page retriever to get the page
	    try {
		currentPage = webpr.getHTMLPage(link);
	    }
	    catch (PathDisallowedException e) {
		System.out.println("Robots excluded from access to" + link + ": " + e);
		continue;
	    }
	    if (currentPage.empty()) {
		System.out.println("No Page Found");
		continue;
	    }
	    if (currentPage.indexAllowed()) {
		count++;
		System.out.println("Indexing(" + count + "): " + link);
		processPage(currentPage);
	    }
	    if (count < maxCount) {
		List newLinks = getNewLinks(currentPage);
		// System.out.println("Adding the following links" + newLinks);
		// Add new links to end of queue
		linksToVisit.addAll(newLinks);
	    }
	}
    }

    /** This version extracts anchored links, stores file name for this
     *  URL in the urlMap, and adds initial link anchor text to the
     *  bottom of the file.
     *
     * @param page An <code>HTMLPage</code> that contains the page to
     * index.  */
    protected void processPage(HTMLPage page) {
	// Extract links in order to change them to absolute URL's
	new AnchoredLinkExtractor(page).extractLinks();
	String fileName = "P" +  MoreString.padWithZeros(count,(int)Math.floor(MoreMath.log(maxCount, 10)) + 1);
	page.writeAbsolute(saveDir, fileName);
	urlMap.put(page.getLink(), fileName);
	addAnchorText(fileName, (AnchoredLink)page.getLink(), true);
    }

    /** Add anchor text of the link to the end of a page file */
    protected void addAnchorText(String fileName, AnchoredLink link, boolean firstLink) {
	if (link.getAnchorText() != null)
	    try {
		String fullName = new File(saveDir, fileName + ".html").toString();
		System.out.println("Adding anchor text to " + fileName);
		// The true flag to FileWriter tells it to append to file
		PrintWriter out = new PrintWriter(new FileWriter(fullName, true));
		if (firstLink)
		    out.println("\n<hr><!-- Added anchor text follows -->");
		out.println("<p>" + link.getAnchorText());
		out.close();
	    }
	    catch (IOException e) {
		System.err.println("AnchoredSpider.addAnchorText(" + fileName + "): " + e);
	    }
    }

    /** Spider the web according to the following command options,
     * but only below the start URL directory. This time storing
     * anchor text to end of pages.
     * <ul> 
     * <li>-safe : Check for and obey robots.txt and robots META tag 
     * directives.</li> 
     * <li>-d &lt;directory&gt; : Store indexed files in &lt;directory&gt.</li>
     * <li>-c &lt;count&gt; : Store at most &lt;count&gt; files.</li> 
     * <li>-u &lt;url&gt; : Start at &lt;url&gt;.</li>
     * <li>-slow : Pause briefly before getting a page.  This can be 
     * useful when debugging.
     * </ul>
     */
    public static void main(String args[]) {
	new AnchoredSpider().go(args);
    }

}
