package ir.webutils;
import java.util.*;
import java.io.*;
import ir.utilities.*;

/**
 * HTMLPage is a representation of information about a web
 * page.
 *
 * @author Ted Wild and Ray Mooney */

public class HTMLPage {

    /** The original link to this page */
    protected final Link link;

    /** The text of the page */
    protected final String text;

    /** The links on this page */
    protected List outLinks;

    /** Copy of the text with relative links replaced by
     * absolute ones */
    protected String absoluteText;

    /** Constructs an <code>HTMLPage</code> with the given link and text.  
     *
     * @param link <code>Link</code> object to the given page.
     *
     * @param text The text of the page.
     */
    public HTMLPage(Link link, String text) {
	this.link = link;
	this.text = text;
    }

    /** Returns the full text of this page.  None of the HTML is
     * stripped out.
     *
     * @return The text of the this page.  */
    public String getText() {
	return text;
    }

    /** Returns the <code>Link</code> object that was used to access
     * this page.
     *
     * @return The <code>Link</code> object that was used to access
     * this page.  */
    public Link getLink() {
	return link;
    }

    /** Set of the outLinks for this page to given list */
    public void setOutLinks(List links) {
	outLinks = links;
    }

    /** Get the list of out links from this page. */
    public List getOutLinks() {
	return outLinks;
    }

    /** Set the absolute link version  of this page */
    public void setAbsoluteText(String text) {
	absoluteText = text;
    }

    /** Get the absolute link version  of this page */
    public String getAbsoluteText() {
	return absoluteText;
    }

    /** Clients should always call this method before indexing an HTML
     * page if they want to obey the "NOINDEX" directive in the Robots
     * META tag.  Always returns <code>true</code> in default implementation.
     * 
     * @return <code>true</code> iff. the page can be indexed.  Always
     * returns <code>true</code> in the default implementation. */
    public boolean indexAllowed() {
	return true;
    }

    /** Returns true if the page is empty or a 404 error. */
    public boolean empty() {
	if(text.equals("") ||
	   MoreString.indexOfIgnoreCase(text, "<title>404 Not Found") >= 0)
	  return true;
	return false;
    }

    /** Writes web page to a file with absolute links and a comment with the
     * original URL.
     *
     * @param dir The directory to store the file in.
     *
     * @param name The name of the file.  */
    public void writeAbsolute(File dir, String name) {
        try {
            PrintWriter out = new PrintWriter(new FileWriter(new File(dir, name + ".html")));
            out.println("<!-- Original URL=\"" + link.getURL() + "\"-->");
            out.print(absoluteText);
            out.close();
        }
        catch (IOException e) {
            System.err.println("HTMLPage.writeAbsolute(): " + e);
        }
    }

}
