package ir.webutils;
import java.net.*;
import java.io.*;
import javax.swing.text.*;
import javax.swing.text.html.*;
import javax.swing.text.html.parser.*;
import java.util.*;
import ir.utilities.*;

/**
 * LinkExtractor defines a callback that extracts the links from an
 * HTML document and provides functionality to parse a document.  The
 * extracted links are absolute.  Uses the HTML parser in Java Swing
 * to parse the document and find links and replace them with
 * absolute URL's (instead of relative ones).
 *
 * @author Ted Wild and Ray Mooney*/

public class LinkExtractor extends HTMLEditorKit.ParserCallback {

    /** The current list of extracted links */
    protected List links;

    /** The page from which to extract links */
    protected HTMLPage page;

    /** The URL for this page */
    protected URL url;

    /** Copy of the text of page with absolute links */
    protected StringBuffer absoluteCopy = null;

    /** Create an link extractor for the given page */
    public LinkExtractor(HTMLPage page) {
	this.links = new LinkedList();
	this.page = page;
	this.url = addEndSlash(page.getLink().getURL());
    }

    /** If URL looks like a directory rather than a file, then
     * add a "/" at the end so that it acts as a proper base URL
     * for completing URLs in this page */
    protected static URL addEndSlash(URL url) {
	String fileName = url.getPath();
	if (MoreString.fileExtension(fileName).equals(""))
	    try {
		return new URL(url.toString() + "/");
	    }
	    catch (MalformedURLException e) {
		System.err.println("LinkExtractor: " + e);
	    }
	return url;
    }
	    
    /** 
     * Executed when a block of text is encountered. Just stores
     * it in the absolute copy.
     *
     * @param text A <code>char</code> array representation of the
     * text.
     *
     * @param position The position of the text in the document.  */
    public void handleText(char[] text, int position) {
	absoluteCopy.append(text);
    }

    /**
     * Executed when an opening HTML tag is found in the document.
     * Note that this method only handles tags that also have a
     * closing tag. Catches "a" tags and adds links for them
     * (after completing them) Also stores completed URL in
     * the absolute copy.
     *
     * @param tag The tag that caused this function to be executed.
     * @param attributes The attributes of <code>tag</code>.
     * @param position The start of the tag in the document.  If the
     * tag is implied (filled in by the parser but not actually
     * present in the document) then <code>position</code> will
     * correspond to that of the next encountered tag.  */
    public void handleStartTag(HTML.Tag tag, MutableAttributeSet attributes, int position) {
	
	if (tag == HTML.Tag.A) {
	    addLink(attributes, HTML.Attribute.HREF);
	}
	appendTag(absoluteCopy, tag, attributes);
    }


    /** Write this tag with attributes out to the buffer */
    public static void appendTag(StringBuffer buffer, HTML.Tag tag, MutableAttributeSet attributes) {
	buffer.append("<" + tag);
	for (Enumeration e = attributes.getAttributeNames(); e.hasMoreElements(); ) {
	    Object attr = e.nextElement();
	    buffer.append(" " + attr + "=");
	    buffer.append("\"" + attributes.getAttribute(attr) + "\"");
	}
	buffer.append(">");
    }
	
    /**
     * Executed when a closing HTML tag is found in the document.
     * Note that the parser may add "implied" closing tags.  For
     * example, the default parser adds closing &lt;p&gt; tags.
     * This just writes end tag out to absolute copy.
     *
     * @param tag The tag found.
     *
     * @param position The position of the tag in the document.  */
    public void handleEndTag(HTML.Tag tag, int position) {
	absoluteCopy.append("</" + tag + ">");
    }

    /**
     * Executed when an HTML tag that has no closing tag is found in
     * the document. Adds link for FRAME's and writes them out to the
     * absolute copy.
     *
     * @param tag The tag that caused this function to be executed.
     * @param attributes The attributes of <code>tag</code>.
     * @param position The start of the tag in the document.  If the
     * tag is implied (filled in by the parser but not actually
     * present in the document) then <code>position</code> will
     * correspond to that of the next encountered tag.  */
    public void handleSimpleTag(HTML.Tag tag, MutableAttributeSet attributes, int position) {
	
	if (tag == HTML.Tag.FRAME) {
	    addLink(attributes, HTML.Attribute.SRC);
	}
	appendTag(absoluteCopy, tag, attributes);
    }

    /** Extracts links from the given page.  This method constructs a
     * parser and registers <code>this</code> as the callback.
     *
     * @return A list of <code>Link</code> objects containing the
     * links found on this page.  The links will all be absolute
     * links. */
    public List extractLinks() {
	HTMLParserMaker kit = new HTMLParserMaker();
	HTMLEditorKit.Parser parser = kit.getParser();
	StringReader reader = new StringReader(page.getText());
	absoluteCopy = new StringBuffer();
	// Swing HTML parser will execute callback routines and thereby
	// extract links and construct an absolute copy of the page
	try {
	    parser.parse(reader, this, true);
	}
	catch (ChangedCharSetException e) {
	    System.err.println("LinkExtractor.extractLinks(): " + e);
	}
	catch (IOException e) {
	    System.err.println("LinkExtractor.extractLinks(): " + e);
	}
	// Set out-links and absolute text for the page
	page.setOutLinks(this.links);
	page.setAbsoluteText(absoluteCopy.toString());
	return this.links;
    }

    /** Retrieves a link from an attribute set and completes it against
     * the base URL.
     *
     * @param attributes The attribute set.
     * @param attr The attribute that should be treated as a URL.  For
     * example, <code>attr</code> should be
     * <code>HTML.Attribute.HREF</code> if <code>attributes</code> is
     * from an anchor tag. */
    protected void addLink(MutableAttributeSet attributes, HTML.Attribute attr) {
	if (attributes.isDefined(attr)) {
	    String link = (String)attributes.getAttribute(attr);
	    try {
		URL completeURL =  new URL(this.url, link);
		if (!link.startsWith("#")) 
		    this.links.add(new Link(completeURL));
		// Change attribute to have complete URL for
		// including in absolute copy
		attributes.removeAttribute(attr);
		attributes.addAttribute(attr, completeURL.toString());
	    }
	    catch (MalformedURLException e) {
		System.err.println("LinkExtractor: " + e);
		// e.printStackTrace(System.err);
	    }
	}
    }


}// LinkExtractor
